from django.db import models
from django.contrib.contenttypes.models import ContentType


__all__ = 'Scope',


class Scope(models.Model):
    class Meta:
        abstract = True

    @classmethod
    def get_content_type(cls):
        return ContentType.objects.get_for_model(cls)



from functools import cached_property
from typing import Callable, Dict, List, Optional, Tuple, Type, Union
from collections import OrderedDict
from django.db import models
from django.utils.translation import pgettext_lazy
from px_domains import raw
from px_domains.registry import DomainPartType, T, RegistryNestResult
from px_access_scopes import ScopeDomain as BaseDomain, ScopeRegistry as BaseRegistry


__all__ = (
    'ScopeDomain',
    'ScopeEnum',
    'ScopeRegistry',
    'auto',
    'raw',
)

EnumDefinition = Dict[str, Union[Tuple[str, str], str]]
TVal = Tuple[str, str, str]
EnumField = Union[TVal, Tuple[str, str]]
EnumFields = List[EnumField]


def get_tuple(values: EnumField) -> TVal:
    if len(values) == 3:
        return values

    return (values + (None, None, None))[:3]


def get_labeled(value, label = None):
    if label is None:
        return value

    return value, label


class ScopeDomain(BaseDomain):
    _delimiter: str = ':'

    @cached_property
    def permission(self):
        from .utils import get_permission_string

        return get_permission_string(self)


class ScopeLabelDomain(BaseDomain):
    _delimiter: str = pgettext_lazy('pxd_access_scopes', ' : ')


class ScopeEnum(ScopeDomain, models.Choices):
    pass


# FIXME:
# Due to an strange work of django's enums auto will be a str
# instead of object.
auto: ScopeDomain = 'auto-' + str(object())


class ScopeRegistry(BaseRegistry):
    """Django-specific domain registry for an access scopes.

    Examples:
        Creating registry root

        >>> root = ScopeRegistry.create_root('ROOT')
        >>>
        >>> @root.nest('SOME')
        >>> class Tokens(Enum):
        >>>     AUTO1 = auto
        >>>     AUTO2 = auto
        >>>     AUTO3 = raw('AUTO3')
        >>>     FIXED = 'some'

        Can be nested in every possible way.

        >>> @Tokens.nest('NESTED')
        >>> class Nested(Enum):
        >>>     AUTO = auto

        `Tokens.AUTO1` will be: `'ROOT::SOME::AUTO1'`
    """

    _Domain = ScopeDomain
    _LabelDomain = ScopeLabelDomain
    _Enum = ScopeEnum
    _auto = auto
    _labels: OrderedDict


    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)

        self._labels = OrderedDict()

    def _register_set(self, domain: ScopeDomain, enum: ScopeEnum):
        super()._register_set(domain, enum)

        for member in enum:
            self._root._labels[member] = member.label

    def _resolve_field_dict(self, domain: ScopeDomain, fields: EnumFields) -> EnumDefinition:
        values = {
            key: get_labeled(
                self._resolve_field_value(domain, key, value),
                label
            )
            for key, value, label in (get_tuple(t) for t in fields)
        }

        return values

    def _resolve_choices(self, domain: ScopeDomain, definition: Type[models.Choices]) -> EnumDefinition:
        return self._resolve_field_dict(domain, [
            (key, value.value, definition._value2label_map_.get(value.value))
            for key, value in definition._member_map_.items()
        ])

    def _resolve_definition(self, domain: ScopeDomain, definition: object) -> EnumDefinition:
        if issubclass(definition, models.Choices):
            return self._resolve_choices(domain, definition)

        return super()._resolve_definition(domain, definition)

    def _get_choices_label(self, domain: str):
        labels = self._root._labels
        ancestors = self._lookup_ancestors(domain)

        return self._LabelDomain.from_definition([str(labels.get(x)) for x in ancestors])

    def get_choices(self, domains: List[ScopeDomain]):
        empty = [(None, self.__empty__)] if hasattr(self, '__empty__') else []

        return empty + [
            (domain, self._get_choices_label(domain)) for domain in domains
        ]

    def nest(
        self,
        name: DomainPartType,
        label: Optional[str] = None
    ) -> Callable[[T], RegistryNestResult[T]]:
        """Decorator for new set registration."""
        internal = super().nest(name)

        def decorator(*args) -> RegistryNestResult[T]:
            instance = internal(*args)
            instance._root._labels[instance._domain] = label if label is not None else instance._domain._definition[-1]

            return instance

        return decorator

    @classmethod
    def create_root(
        cls,
        domain: DomainPartType,
        label: Optional[str] = None
    ) -> 'ScopeRegistry':
        """Creates new root with the domain as a base."""

        instance = super(ScopeRegistry, cls).create_root(domain)
        instance._root._labels[instance._domain] = label if label is not None else instance._domain

        return instance



from typing import List, Union

from .scopes import ScopeRegistry
from .aggregates import Aggregate, Aggregates


__all__ = 'registries', 'aggregates'

registries: List[ScopeRegistry] = []
aggregates: List[Union[Aggregate, Aggregates]] = []



from typing import Optional, Tuple, Union
from px_access_scopes import Aggregate as BaseAggregate, Aggregates

from .scopes import ScopeDomain, ScopeEnum


__all__ = 'Aggregate', 'Aggregates',

# Group identifier must be in a range of {2000 - 4000}
# to avoid id collisions with an existing database ids.
# It will be better to initially create aggregates with larger ids
# and the newer they are - the lower they become.
# It's because of id autoincrement database mechanics.
GROUP_IDS_RANGE = (2000, 5000)
AggregateValue = Union[ScopeDomain, ScopeEnum]


class Aggregate(BaseAggregate):
    """Set with additional config parameters."""

    _group_ids_range: Tuple[int, int] = GROUP_IDS_RANGE
    group_id: int
    name: str
    verbose_name: str

    def __init__(
        self,
        group_id: int,
        name: Optional[str] = None,
        verbose_name: Optional[str] = None,
        *args
    ) -> None:
        lowest, highest = self._group_ids_range
        assert lowest <= group_id <= highest, (
            f'Group identifier must be in a range of {lowest}-{highest} '
            'to avoid id collisions with an existing database ids.'
        )

        super().__init__(name=name or group_id, *args)
        self.group_id = group_id
        self.verbose_name = verbose_name or self.name



from django.utils.module_loading import autodiscover_modules
from django.utils.module_loading import import_string

from .conf import settings
from .globals import registries, aggregates


__all__ = 'autodiscover',


def autodiscover():
    autodiscover_modules(settings.AUTOLOAD_MODULE)

    for registry in settings.REGISTRIES or []:
        registries.append(import_string(registry))

    for aggregate in settings.AGGREGATES or []:
        aggregates.append(import_string(aggregate))



from functools import lru_cache
from typing import Iterator, List, Union

from .aggregates import Aggregate, Aggregates


__all__ = (
    'get_permission_app_label',
    'get_permission_string',
    'flatten_aggregates',
)


@lru_cache
def get_permission_app_label():
    from .models import Scope

    return Scope._meta.app_label


def get_permission_string(key: str):
    return get_permission_app_label() + '.' + key


def flatten_aggregates(
    aggregates: List[Union['Aggregate', 'Aggregates']]
) -> Iterator['Aggregate']:
    return (
        a
        for agg in aggregates
        for a in (
            (getattr(agg, name) for name in agg._item_names)
            if isinstance(agg, type) and issubclass(agg, Aggregates)
            else (agg,)
        )
    )



from typing import Iterable, Set, Type, TypeVar, Union, TYPE_CHECKING
from enum import Enum

if TYPE_CHECKING:
    from .scopes import ScopeDomain


T = TypeVar('T')
D = TypeVar('D', bound='ScopeDomain')
ScopeType = Union['ScopeDomain', Enum, str]


def to_set(value: Iterable[T]) -> Set[T]:
    return value if isinstance(value, set) else set(value)


def normalize_scope(scope: ScopeType, cls: Type[D] = 'ScopeDomain') -> D:
    if isinstance(scope, Enum):
        scope = scope.value

    if not isinstance(scope, cls):
        scope = cls.parse(scope)

    return scope



from typing import List
from django.contrib.auth.models import Permission

from ..scopes import ScopeRegistry
from ..models import Scope


__all__ = 'generate_permissions',


def generate_permissions(registry: ScopeRegistry) -> List[Permission]:
    keys = set(registry.lookup_descendants())
    content_type = Scope.get_content_type()
    existing = (
        Permission.objects
        .filter(codename__in=keys)
        .values_list('codename', flat=True)
    )
    choices = registry.get_choices(keys.difference(existing))

    return Permission.objects.bulk_create([
        Permission(codename=value, name=label, content_type=content_type)
        for value, label in choices
    ])



from typing import Optional, Sequence, Set, Type

from ..scopes import ScopeRegistry, ScopeDomain
from ..utils import normalize_scope, ScopeType
from ..exceptions import CheckExit
from .base import ScopesChecker
from .checkers import run_checkers, ScopesCheckerType


__all__ = 'ScopesCheckRunner',


class ScopesCheckRunner(ScopesChecker):
    """Root runner for multiple checkers.

    Instance might be used as a simple checker inside another runner
    but with small exceptions:

    - It transforms passed scopes into an appropriate `.Domain` instances.
    - It handles `CheckExit` and returns `False` in that case.
    - It adds `registry` to a provided kwargs, by replasing any passed through.
    """

    Domain: Type[ScopeDomain] = ScopeDomain
    registry: Optional[ScopeRegistry]
    checkers: Sequence[ScopesCheckerType]

    def __init__(
        self,
        checkers: Sequence[ScopesCheckerType],
        registry: Optional[ScopeRegistry] = None
    ):
        assert len(checkers) != 0, 'There must be at least one checker.'

        self.checkers = checkers
        self.registry = registry

    def _normalize_scope(self, scope: ScopeType):
        return normalize_scope(scope, self.Domain)

    def _run_checkers(
        self,
        check_scopes: Sequence[ScopeType],
        kwargs: dict = {}
    ) -> bool:
        kwargs['registry'] = self.registry

        return run_checkers(check_scopes, self.checkers, kwargs)

    def has(self, check_scopes: Sequence[ScopeType], *_, **kwargs) -> bool:
        check_scopes = {
            self._normalize_scope(scope) for scope in check_scopes
        }

        try:
            return self._run_checkers(check_scopes, kwargs)
        except CheckExit:
            return False



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy
from django.dispatch import receiver
from django.db.models.signals import post_migrate

from .discover import autodiscover


__all__ = ('AccessScopesConfig',)


class AccessScopesConfig(AppConfig):
    name = 'px_access_scopes.contrib.django'
    label = 'pxd_access_scopes_django'
    verbose_name = pgettext_lazy('pxd_access_scopes', 'Access scopes')

    def ready(self):
        from . import subscriptions

        autodiscover()

        receiver(post_migrate, sender=self)(subscriptions.generate_data_on_migrate)



from typing import Iterable, List, Type, Union

from .scopes import ScopeDomain, ScopeEnum, auto as auto_scopes


__all__ = (
    'Aggregate',
    'auto_aggregate',
    'AggregatesMeta',
    'BaseAggregates',
    'Aggregates',
)


AggregateValue = Union[ScopeDomain, ScopeEnum]


class Aggregate(set):
    """Set with 'name' attribute. Nothing more."""

    name: str

    def __init__(
        self,
        name: str,
        iterable: Iterable[AggregateValue] = ()
    ) -> None:
        super().__init__(iterable)
        self.name = name

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f'<{self.name}: {super().__repr__()}>'


auto = auto_aggregate = Aggregate('auto')


class AggregatesMeta(type):
    AGGREGATE_ATTRIBUTE = '_Aggregate'
    VALUE_GENERATOR_ATTRIBUTE = '_generate_next_value_'

    def __new__(metacls, cls, bases, classdict):
        value_generator = metacls._resolve_cls_attribute(
            bases, classdict, metacls.VALUE_GENERATOR_ATTRIBUTE
        )
        names = []
        Class = metacls._resolve_cls_attribute(
            bases, classdict, metacls.AGGREGATE_ATTRIBUTE
        )

        if Class is None:
            raise TypeError(
                f'No attribute `{metacls.AGGREGATE_ATTRIBUTE}` defined '
                'in class or it\'s bases.'
            )

        for key, value in classdict.items():
            if value is auto or value is auto_scopes:
                assert value_generator is not None, (
                    f'Add `{metacls.VALUE_GENERATOR_ATTRIBUTE}` method to '
                    'a class `{cls}` to be able to use auto generation.'
                )

                value = classdict[key] = value_generator(cls=Class, key=key)

            if isinstance(value, Class):
                names.append(key)

        classdict['_item_names'] = names

        instance = super().__new__(
            metacls, cls, bases, classdict
        )

        return instance

    @classmethod
    def _resolve_cls_attribute(metacls, bases, classdict, attr: str):
        if attr in classdict:
            return classdict[attr]

        for chain in bases:
            for base in chain.__mro__:
                if hasattr(base, attr):
                    return getattr(base, attr)


class BaseAggregates(metaclass=AggregatesMeta):
    _Aggregate: Type[Aggregate] = Aggregate
    _item_names: List[str]


class Aggregates(BaseAggregates):
    """Special Enum-like class that can autogenerate aggregate attributes."""

    def _generate_next_value_(cls, key: str):
        return cls(key)



"""Django application for a little bit easier access scopes integration.
"""

from .scopes import *
from .aggregates import *

default_app_config = 'px_access_scopes.contrib.django.apps.AccessScopesConfig'



from typing import List
from django.contrib.auth.models import Permission, Group
from px_access_scopes.aggregates import Aggregate

from ..aggregates import Aggregate
from ..models import Scope


__all__ = 'generate_group',


def generate_group(aggregate: Aggregate) -> Group:
    group, _ = Group.objects.get_or_create(
        pk=aggregate.group_id, defaults={'name': aggregate.verbose_name}
    )
    group.permissions.set(Permission.objects.filter(
        codename__in=aggregate, content_type=Scope.get_content_type()
    ))

    return group
